# Configuration file for Sage's builtin Jupyter notebook server

# Disable XSRF checking to fix Thebe. See
# * https://trac.sagemath.org/ticket/22458
# * https://github.com/oreillymedia/thebe/issues/93

c.NotebookApp.disable_check_xsrf = True

# send2trash sometimes doesn't work, so disable that for now
# See https://github.com/jupyter/notebook/issues/3249

c.FileContentsManager.delete_to_trash = False



import os
from sage.all import save
from sage.env import SAGE_SHARE
from sage.misc.misc import sage_makedirs

install_root = os.path.join(SAGE_SHARE, 'odlyzko')
target = os.path.join(install_root, 'zeros.sobj')

if __name__ == '__main__':
    sage_makedirs(install_root)
    print("Creating Odlyzko database.")
    F = [float(x) for x in open("src/zeros6").readlines()]
    save(F, target)



from configparser import ConfigParser
import pkgconfig
import os

config = ConfigParser()

config.add_section('directories')
config.set('directories', 'basedirlist', os.environ['SAGE_LOCAL'])

config.add_section('libs')
config.set('libs', 'system_freetype', 'True')
if pkgconfig.installed('qhull', '>= 7.2.0'):
    config.set('libs', 'system_qhull', 'True')
# lto is problematic if we mix libraries from the OS with our own libraries,
# which are not necessarily compiled with the same gcc version
# https://trac.sagemath.org/ticket/27754
config.set('libs', 'enable_lto', 'False')

#####################################################################
# Sage code -- all this code just sets the graphical_backend variable.
# If True, that means we try to build GUI's; otherwise, we definitely
# will not even try, even if we could.  See trac #5301.
#####################################################################

print("NOTE: Set SAGE_MATPLOTLIB_GUI to anything but 'no' to try to build the Matplotlib GUI.")

graphical_backend='False'
if os.environ.get('SAGE_MATPLOTLIB_GUI', 'no').lower() != 'no':
    graphical_backend = 'auto'

if graphical_backend == 'auto':
    print("Building graphical backends.  WARNING: This may causing some annoying and confusing behavior")
    print("when using Sage + pylab, at least on OS X.")
else:
    print("Not building any matplotlib graphical backends.")

config.add_section('gui_support')
for backend in ('gtk', 'gtkagg', 'tkagg', 'wxagg', 'macosx', 'windowing'):
    config.set('gui_support', backend,  graphical_backend)

with open('src/setup.cfg', 'w') as configfile:
    config.write(configfile)



# -*- coding: utf-8 -*-
"""
    inventory builder
    ~~~~~~~~~~~~~~~~~

    A customized HTML builder which only generates intersphinx "object.inv"
    inventory files and pickle files. The documentation files are not written.
"""
from sphinx.builders.html import StandaloneHTMLBuilder
from os import path
import shutil


class InventoryBuilder(StandaloneHTMLBuilder):
    """
    A customized HTML builder which only generates intersphinx "object.inv"
    inventory files and pickle files. The documentation files are not written.
    """
    name = "inventory"
    format = "inventory"
    epilog = "The inventory files are in %(outdir)s."

    def get_outdated_docs(self):
        from sphinx.builders.html import BuildInfo
        old = BuildInfo()
        try:
            with open(path.join(self.outdir, '.buildinfo')) as fp:
                old = BuildInfo.load(fp)
        except ValueError:
            self.warn('unsupported build info format in %r, building all' %
                      path.join(self.outdir, '.buildinfo'))
        except Exception:
            pass
        if self.build_info != old:
            for docname in self.env.found_docs:
                yield docname
            return

        if self.templates:
            template_mtime = self.templates.newest_template_mtime()
        else:
            template_mtime = 0
        for docname in self.env.found_docs:
            if docname not in self.env.all_docs:
                yield docname
                continue
            targetname = path.join(self.outdir, 'objects.inv')
            try:
                targetmtime = path.getmtime(targetname)
            except Exception:
                targetmtime = 0
            try:
                srcmtime = max(path.getmtime(self.env.doc2path(docname)),
                               template_mtime)
                if srcmtime > targetmtime:
                    yield docname
            except EnvironmentError:
                # source doesn't exist anymore
                pass

    def write_doc(self, docname, doctree):
        """
        Don't write any doc
        """

    def finish(self):
        """
        Only write the inventory files.
        """
        self.write_buildinfo()
        self.dump_inventory()

    def removed_method_error(self):
        """
        Raise an error if this method is called.

        This is just for making sure that some writer methods are indeed
        deactivated.
        """
        raise RuntimeError("This function shouldn't be called in \"%s\" builder"%(self.name))

    def cleanup(self):
        """
        Remove the '_static' directory.

        This directory is unnecessary for the inventory build, but it
        may be created by the graphviz extension. Its presence will
        break the docbuild later on, so remove it.
        """
        if path.isdir(path.join(self.outdir, '_static')):
            shutil.rmtree(path.join(self.outdir, '_static'))


    copy_image_files = removed_method_error
    copy_download_files = removed_method_error
    copy_static_files = removed_method_error
    handle_finish = removed_method_error


def setup(app):
    app.add_builder(InventoryBuilder)
    return {'parallel_read_safe': True} 



import sys

def compiler_directives(profile: bool):
    """
    Return a list of Cython directives used for compilation.
    """
    return dict(
        # Do not generate __reduce__ methods
        auto_pickle=False,
        # Do not create __test__ dictionary automatically from docstrings
        autotestdict=False,
        # Do not check for division by 0 (this is about 35% quicker than with check)
        cdivision=True,
        # Embed a textual copy of the call signature in the docstring (to support tools like IPython)
        embedsignature=True,
        fast_getattr=True,
        # Use Python 3 (including source code semantics) for module compilation
        language_level="3str",
        # Enable support for late includes (make declarations in Cython code available to C include files)
        preliminary_late_includes_cy28=True,
        # Add hooks for Python profilers into the compiled C code
        profile=profile,
    )

def compile_time_env_variables():
    """
    Return a list of environmental variables used for compilation.
    """
    return dict(
        PY_PLATFORM=sys.platform,
        # The following two constants are here only for backwards compatibility of user packages
        PY_VERSION_HEX=sys.hexversion,
        PY_MAJOR_VERSION=sys.version_info[0]
    )



# -*- coding: utf-8 -*-
"""
Test Sage Package Handling
"""

# ****************************************************************************
#       Copyright (C) 2015 Volker Braun <vbraun.name@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

import unittest
from sage_bootstrap.package import Package
from sage_bootstrap.tarball import Tarball


class PackageTestCase(unittest.TestCase):

    maxDiff = None

    def test_package(self):
        pkg = Package('pari')
        self.assertTrue(pkg.name, 'pari')
        self.assertTrue(pkg.path.endswith('build/pkgs/pari'))
        self.assertEqual(pkg.tarball_pattern, 'pari-VERSION.tar.gz')
        self.assertEqual(pkg.tarball_filename, pkg.tarball.filename)
        self.assertTrue(pkg.tarball.filename.startswith('pari-') and
                        pkg.tarball.filename.endswith('.tar.gz'))
        self.assertTrue(pkg.tarball.filename.startswith('pari-') and
                        pkg.tarball.filename.endswith('.tar.gz'))
        self.assertTrue(isinstance(pkg.tarball, Tarball))

    def test_all(self):
        pari = Package('pari')
        self.assertTrue(pari in Package.all())



"""
Verify that the ATLAS enums are in sync with the spkg-install enums

TESTS::

    sage: check_enums(sample_print_enums_output)
"""

from __future__ import print_function

# constants from src/ATLAS/CONFIG/include/atlconf.h
# Note: must be lists, not tuples, for Python-2.4 support

ATLAS_OSTYPE = (  # static char *osnam
    'UNKNOWN', 'Linux', 'SunOS', 'SunOS4', 'OSF1', 'IRIX', 'AIX',
    'Win9x', 'WinNT', 'Win64', 'HPUX', 'FreeBSD', 'OSX')

ATLAS_MACHTYPE = (  # static char *machnam
    'UNKNOWN', 'POWER3', 'POWER4', 'POWER5', 'PPCG4', 'PPCG5',
    'POWER6', 'POWER7', 'POWERe6500', 'IBMz9', 'IBMz10', 'IBMz196',
    'x86x87', 'x86SSE1', 'x86SSE2', 'x86SSE3',
    'P5', 'P5MMX', 'PPRO', 'PII', 'PIII', 'PM', 'CoreSolo',
    'CoreDuo', 'Core2Solo', 'Core2', 'Corei1', 'Corei2', 'Corei3',
    'Atom', 'P4', 'P4E',
    'Efficeon', 'K7', 'HAMMER', 'AMD64K10h', 'AMDLLANO', 'AMDDOZER','AMDDRIVER',
    'UNKNOWNx86', 'IA64Itan', 'IA64Itan2',
    'USI', 'USII', 'USIII', 'USIV', 'UST1', 'UST2', 'UnknownUS',
    'MIPSR1xK', 'MIPSICE9', 'ARMv6', 'ARMv7')

ATLAS_ISAEXT = (  # static char *ISAXNAM
    'None', 'VSX', 'AltiVec', 'AVXMAC', 'AVXFMA4', 'AVX', 'SSE3', 'SSE2', 'SSE1',
    '3DNow', 'NEON')



# for doctesting purposes
sample_print_enums_output = \
"""
Architectural enums (Config's enum MACHTYPE):
     0 = 'UNKNOWN'
     1 = 'POWER3'
     2 = 'POWER4'
     3 = 'POWER5'
     4 = 'PPCG4'
     5 = 'PPCG5'
     6 = 'POWER6'
     7 = 'POWER7'
     8 = 'POWERe6500'
     9 = 'IBMz9'
    10 = 'IBMz10'
    11 = 'IBMz196'
    12 = 'x86x87'
    13 = 'x86SSE1'
    14 = 'x86SSE2'
    15 = 'x86SSE3'
    16 = 'P5'
    17 = 'P5MMX'
    18 = 'PPRO'
    19 = 'PII'
    20 = 'PIII'
    21 = 'PM'
    22 = 'CoreSolo'
    23 = 'CoreDuo'
    24 = 'Core2Solo'
    25 = 'Core2'
    26 = 'Corei1'
    27 = 'Corei2'
    28 = 'Corei3'
    29 = 'Atom'
    30 = 'P4'
    31 = 'P4E'
    32 = 'Efficeon'
    33 = 'K7'
    34 = 'HAMMER'
    35 = 'AMD64K10h'
    36 = 'AMDLLANO'
    37 = 'AMDDOZER'
    38 = 'AMDDRIVER'
    39 = 'UNKNOWNx86'
    40 = 'IA64Itan'
    41 = 'IA64Itan2'
    42 = 'USI'
    43 = 'USII'
    44 = 'USIII'
    45 = 'USIV'
    46 = 'UST1'
    47 = 'UST2'
    48 = 'UnknownUS'
    49 = 'MIPSR1xK'
    50 = 'MIPSICE9'
    51 = 'ARMv6'
    52 = 'ARMv7'

Operating System enums (Config's enum OSTYPE):
     0 = 'UNKNOWN'
     1 = 'Linux'
     2 = 'SunOS'
     3 = 'SunOS4'
     4 = 'OSF1'
     5 = 'IRIX'
     6 = 'AIX'
     7 = 'Win9x'
     8 = 'WinNT'
     9 = 'Win64'
    10 = 'HPUX'
    11 = 'FreeBSD'
    12 = 'OSX'

Compiler integer defines:
     0 = 'ICC'
     1 = 'SMC'
     2 = 'DMC'
     3 = 'SKC'
     4 = 'DKC'
     5 = 'XCC'
     6 = 'GCC'
     7 = 'F77'


ISA extensions are combined by adding their values together (bitvector):
         none: 1
          VSX: 2
      AltiVec: 4
       AVXMAC: 8
      AVXFMA4: 16
          AVX: 32
         SSE3: 64
         SSE2: 128
         SSE1: 256
        3DNow: 512
         NEON: 1024

"""


def check_enums(print_enums_output):
    """
    Verify that the output of ATLAS print_enums matches our enums

    INPUT:

    - ``print_enums_output`` -- string. The output of the ATLAS
      print_enums utility.
    """
    lines = print_enums_output.splitlines()
    found_MACHTYPE = found_OSTYPE = found_ISAEXT = False
    while len(lines) > 0:
        line = lines.pop(0)
        if line.startswith('Architectural enums'):
            check_enums_ATLAS_MACHTYPE(lines)
            found_MACHTYPE = True
        if line.startswith('Operating System enums'):
            check_enums_ATLAS_OSTYPE(lines)
            found_OSTYPE = True
        if line.startswith('ISA extensions'):
            check_enums_ATLAS_ISAEXT(lines)
            found_ISAEXT = True
    if not (found_MACHTYPE and found_OSTYPE and found_ISAEXT):
        raise RuntimeError('failed to parse the output of print_enums')


def check_enums_ATLAS_MACHTYPE(lines):
    for i, mach_type in enumerate(ATLAS_MACHTYPE):
        got = lines.pop(0).strip()
        expect = "{0} = '{1}'".format(i, mach_type)
        if got != expect:
            raise RuntimeError('ATLAS_MACHTYPE mismatch at position '+str(i)+
                               ': got >>'+got+'<<, expected >>'+expect+'<<')

def check_enums_ATLAS_OSTYPE(lines):
    for i, os_type in enumerate(ATLAS_OSTYPE):
        got = lines.pop(0).strip()
        expect = "{0} = '{1}'".format(i, os_type)
        if got != expect:
            raise RuntimeError('ATLAS_OSTYPE mismatch at position '+str(i)+
                               ': got >>'+got+'<<, expected >>'+expect+'<<')

def check_enums_ATLAS_ISAEXT(lines):
    for i, isaext in enumerate(ATLAS_ISAEXT):
        got = lines.pop(0).strip()
        if i == 0:
            expect = 'none: 1'
        else:
            expect = "{0}: {1}".format(isaext, 1 << i)
        if got != expect:
            raise RuntimeError('ATLAS_ISAEXT mismatch at position '+str(i)+
                               ': got >>'+got+'<<, expected >>'+expect+'<<')


def make_check_enums():
    """
    Build the print_enums utility and check its output against our
    enums

    You can only call this function when you are in the atlas build
    directory, and only after configuring atlas.
    """
    from subprocess import check_output
    output = check_output('make xprint_enums ; ./xprint_enums', shell=True)
    print(output)
    check_enums(output.decode('ascii'))




import os
import pickle

SAGE_SHARE = os.getenv('SAGE_SHARE')
install_root = os.path.join(SAGE_SHARE, 'conway_polynomials')

def create_db():
    db = {}
    from src import conway_polynomials
    for p, n, v in conway_polynomials:
        if not p in db:
            db[p] = {}
        db[p][n] = v
    if not os.path.exists(install_root):
        os.makedirs(install_root)
    with open(os.path.join(install_root, 'conway_polynomials.p'), 'wb') as f:
        pickle.dump(db, f)

if __name__ == '__main__':
    create_db()



"""
Clean the Install Dir
"""
#*****************************************************************************
#       Copyright (C) 2014 Volker Braun <vbraun.name@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  http://www.gnu.org/licenses/
#*****************************************************************************


import os
import importlib.util

from sage_setup.find import installed_files_by_module, get_extensions



def _remove(file_set, module_base, to_remove):
    """
    Helper to remove files from a set of filenames.

    INPUT:

    - ``file_set`` -- a set of filenames.

    - ``module_base`` -- string. Name of a Python package/module.

    - ``to_remove`` -- list/tuple/iterable of strings. Either
      filenames or extensions (starting with ``'.'``)

    OUTPUT:

    This function does not return anything. The ``file_set`` parameter
    is modified in place.

    EXAMPLES::

        sage: files = set(['a/b/c.py', 'a/b/d.py', 'a/b/c.pyx'])
        sage: from sage_setup.clean import _remove
        sage: _remove(files, 'a.b', ['c.py', 'd.py'])
        sage: files
        {'a/b/c.pyx'}

        sage: files = set(['a/b/c.py', 'a/b/d.py', 'a/b/c.pyx'])
        sage: _remove(files, 'a.b.c', ['.py', '.pyx'])
        sage: files
        {'a/b/d.py'}
    """
    path = os.path.join(*module_base.split('.'))

    for filename in to_remove:
        if filename.startswith('.'):
            filename = path + filename
        else:
            filename = os.path.join(path, filename)

        remove = [filename]

        remove.append(importlib.util.cache_from_source(filename))
        file_set.difference_update(remove)


def _find_stale_files(site_packages, python_packages, python_modules, ext_modules, data_files, nobase_data_files=()):
    """
    Find stale files

    This method lists all files installed and then subtracts the ones
    which are intentionally being installed.

    EXAMPLES:

    It is crucial that only truly stale files are being found, of
    course. We check that when the doctest is being run, that is,
    after installation, there are no stale files::

        sage: from sage.env import SAGE_SRC, SAGE_LIB, SAGE_ROOT
        sage: from sage_setup.find import _cythonized_dir
        sage: cythonized_dir = _cythonized_dir(SAGE_SRC)
        sage: from sage_setup.find import find_python_sources, find_extra_files
        sage: python_packages, python_modules, cython_modules = find_python_sources(
        ....:     SAGE_SRC, ['sage', 'sage_setup'])
        sage: extra_files = list(find_extra_files(SAGE_SRC,
        ....:     ['sage', 'sage_setup'], cythonized_dir, []).items())
        sage: from sage_setup.clean import _find_stale_files

    TODO: move ``module_list.py`` into ``sage_setup`` and also check
    extension modules::

        sage: stale_iter = _find_stale_files(SAGE_LIB, python_packages, python_modules, [], extra_files)
        sage: from sage.misc.sageinspect import loadable_module_extension
        sage: skip_extensions = (loadable_module_extension(),)
        sage: for f in stale_iter:
        ....:     if f.endswith(skip_extensions): continue
        ....:     if '/ext_data/' in f: continue
        ....:     print('Found stale file: ' + f)
    """

    PYMOD_EXTS = get_extensions('source') + get_extensions('bytecode')
    CEXTMOD_EXTS = get_extensions('extension')
    INIT_FILES = tuple('__init__' + x for x in PYMOD_EXTS)

    module_files = installed_files_by_module(site_packages, ['sage', 'sage_setup'])

    for mod in python_packages:
        try:
            files = module_files[mod]
        except KeyError:
            # the source module "mod" has not been previously installed, fine.
            continue
        _remove(files, mod, INIT_FILES)
    for mod in python_modules:
        try:
            files = module_files[mod]
        except KeyError:
            continue
        _remove(files, mod, PYMOD_EXTS)
    for ext in ext_modules:
        mod = ext.name
        try:
            files = module_files[mod]
        except KeyError:
            continue
        _remove(files, mod, CEXTMOD_EXTS)

    # Convert data_files to a set
    installed_files = set()
    for dir, files in data_files:
        for f in files:
            installed_files.add(os.path.join(dir, os.path.basename(f)))
    for dir, files in nobase_data_files:
        for f in files:
            installed_files.add(f)

    for files in module_files.values():
        for f in files:
            if f not in installed_files:
                yield f


def clean_install_dir(site_packages, python_packages, python_modules, ext_modules, data_files, nobase_data_files):
    """
    Delete all modules that are **not** being installed

    If you switch branches it is common to (re)move the source for an
    already installed module. Subsequent rebuilds will leave the stale
    module in the install directory, which can break programs that try
    to import all modules. In particular, the Sphinx autodoc builder
    does this and is susceptible to hard-to-reproduce failures that
    way. Hence we must make sure to delete all stale modules.

    INPUT:

    - ``site_packages`` -- the root Python path where the Sage library
      is being installed.

    - ``python_packages`` -- list of pure Python packages (directories
      with ``__init__.py``).

    - ``python_modules`` -- list of pure Python modules.

    - ``ext_modules`` -- list of distutils ``Extension`` classes. The
      output of ``cythonize``.

    - ``data_files`` -- a list of (installation directory, files) pairs,
      like the ``data_files`` argument to distutils' ``setup()``. Only
      the basename of the files is used.

    - ``nobase_data_files`` -- a list of (installation directory, files)
      pairs. The files are expected to be in a subdirectory of the
      installation directory; the filenames are used as is.

    """
    stale_file_iter = _find_stale_files(
        site_packages, python_packages, python_modules, ext_modules, data_files, nobase_data_files)
    for f in stale_file_iter:
        f = os.path.join(site_packages, f)
        print('Cleaning up stale file: {0}'.format(f))
        os.unlink(f)



#*****************************************************************************
#       Copyright (C) 2009 Carl Witty <Carl.Witty@gmail.com>
#       Copyright (C) 2015 Jeroen Demeyer <jdemeyer@cage.ugent.be>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from __future__ import print_function, absolute_import

from .base import StackInterpreter
from ..instructions import (params_gen, instr_funcall_2args, instr_unary,
                            InstrSpec)
from ..memory import MemoryChunk
from ..storage import ty_python
from ..utils import je, reindent_lines as ri


class MemoryChunkPythonArguments(MemoryChunk):
    r"""
    A special-purpose memory chunk, for the generic Python-object based
    interpreter.  Rather than copy the arguments into an array allocated
    in the wrapper, we use the PyTupleObject internals and pass the array
    that's inside the argument tuple.
    """

    def declare_class_members(self):
        r"""
        Return a string giving the declarations of the class members
        in a wrapper class for this memory chunk.

        EXAMPLES::

            sage: from sage_setup.autogen.interpreters import *
            sage: mc = MemoryChunkPythonArguments('args', ty_python)
        """
        return "    cdef int _n_%s\n" % self.name

    def init_class_members(self):
        r"""
        Return a string to be put in the __init__ method of a wrapper
        class using this memory chunk, to initialize the corresponding
        class members.

        EXAMPLES::

            sage: from sage_setup.autogen.interpreters import *
            sage: mc = MemoryChunkPythonArguments('args', ty_python)
            sage: mc.init_class_members()
            u"        count = args['args']\n        self._n_args = count\n"
        """
        return je(ri(8,
            """
            count = args['{{ myself.name }}']
            self._n_args = count
            """), myself=self)

    def setup_args(self):
        r"""
        Handle the arguments of __call__.  Nothing to do.

        EXAMPLES::

            sage: from sage_setup.autogen.interpreters import *
            sage: mc = MemoryChunkPythonArguments('args', ty_python)
            sage: mc.setup_args()
            ''
        """
        return ''

    def pass_argument(self):
        r"""
        Pass the innards of the argument tuple to the interpreter.

        EXAMPLES::

            sage: from sage_setup.autogen.interpreters import *
            sage: mc = MemoryChunkPythonArguments('args', ty_python)
            sage: mc.pass_argument()
            '(<PyTupleObject*>args).ob_item'
        """
        return "(<PyTupleObject*>args).ob_item"


class MemoryChunkPyConstant(MemoryChunk):
    r"""
    A special-purpose memory chunk, for holding a single Python constant
    and passing it to the interpreter as a PyObject*.
    """

    def __init__(self, name):
        r"""
        Initialize an instance of MemoryChunkPyConstant.

        Always uses the type ty_python.

        EXAMPLES::

            sage: from sage_setup.autogen.interpreters import *
            sage: mc = MemoryChunkPyConstant('domain')
            sage: mc.name
            'domain'
            sage: mc.storage_type is ty_python
            True
        """
        super(MemoryChunkPyConstant, self).__init__(name, ty_python)

    def declare_class_members(self):
        r"""
        Return a string giving the declarations of the class members
        in a wrapper class for this memory chunk.

        EXAMPLES::

            sage: from sage_setup.autogen.interpreters import *
            sage: mc = MemoryChunkPyConstant('domain')
            sage: mc.declare_class_members()
            u'    cdef object _domain\n'
        """
        return je(ri(4,
            """
            cdef object _{{ myself.name }}
            """), myself=self)

    def init_class_members(self):
        r"""
        Return a string to be put in the __init__ method of a wrapper
        class using this memory chunk, to initialize the corresponding
        class members.

        EXAMPLES::

            sage: from sage_setup.autogen.interpreters import *
            sage: mc = MemoryChunkPyConstant('domain')
            sage: mc.init_class_members()
            u"        self._domain = args['domain']\n"
        """
        return je(ri(8,
            """
            self._{{ myself.name }} = args['{{ myself.name }}']
            """), myself=self)

    def declare_parameter(self):
        r"""
        Return the string to use to declare the interpreter parameter
        corresponding to this memory chunk.

        EXAMPLES::

            sage: from sage_setup.autogen.interpreters import *
            sage: mc = MemoryChunkPyConstant('domain')
            sage: mc.declare_parameter()
            'PyObject* domain'
        """
        return 'PyObject* %s' % self.name

    def pass_argument(self):
        r"""
        Return the string to pass the argument corresponding to this
        memory chunk to the interpreter.

        EXAMPLES::

            sage: from sage_setup.autogen.interpreters import *
            sage: mc = MemoryChunkPyConstant('domain')
            sage: mc.pass_argument()
            '<PyObject*>self._domain'
        """
        return '<PyObject*>self._%s' % self.name


class PythonInterpreter(StackInterpreter):
    r"""
    A subclass of StackInterpreter, specifying an interpreter over
    Python objects.

    Let's discuss how the reference-counting works in Python-object
    based interpreters.

    There is a simple rule to remember: when executing the code
    snippets, the input variables contain borrowed references;
    you must fill in the output variables with references you own.

    As an optimization, an instruction may set .handles_own_decref; in
    that case, it must decref any input variables that came from the
    stack.  (Input variables that came from arguments/constants chunks
    must NOT be decref'ed!)  In addition, with .handles_own_decref, if
    any of your input variables are arbitrary-count, then you must
    NULL out these variables as you decref them.  (Use Py_CLEAR to do
    this, unless you understand the documentation of Py_CLEAR and why
    it's different than Py_XDECREF followed by assigning NULL.)

    Note that as a tiny optimization, the interpreter always assumes
    (and ensures) that empty parts of the stack contain NULL, so
    it doesn't bother to Py_XDECREF before it pushes onto the stack.
    """

    name = 'py'

    def __init__(self):
        r"""
        Initialize a PythonInterpreter.

        EXAMPLES::

            sage: from sage_setup.autogen.interpreters import *
            sage: interp = PythonInterpreter()
            sage: interp.name
            'py'
            sage: interp.mc_args
            {MC:args}
            sage: interp.chunks
            [{MC:args}, {MC:constants}, {MC:stack}, {MC:code}]
            sage: instrs = dict([(ins.name, ins) for ins in interp.instr_descs])
            sage: instrs['add']
            add: SS->S = 'o0 = PyNumber_Add(i0, i1);'
            sage: instrs['py_call']
            py_call: *->S = '\nPyObject *py_args...CREF(py_args);\n'
        """

        super(PythonInterpreter, self).__init__(ty_python)
        # StackInterpreter.__init__ gave us a MemoryChunkArguments.
        # Override with MemoryChunkPythonArguments.
        self.mc_args = MemoryChunkPythonArguments('args', ty_python)
        self.chunks = [self.mc_args, self.mc_constants, self.mc_stack,
                       self.mc_code]
        self.c_header = ri(0,
            """
            #define CHECK(x) (x != NULL)
            """)

        self.pyx_header = ri(0,
            """\
            from cpython.number cimport PyNumber_TrueDivide
            """)

        pg = params_gen(A=self.mc_args, C=self.mc_constants, D=self.mc_code,
                        S=self.mc_stack)
        self.pg = pg

        instrs = [
            InstrSpec('load_arg', pg('A[D]', 'S'),
                       code='o0 = i0; Py_INCREF(o0);'),
            InstrSpec('load_const', pg('C[D]', 'S'),
                       code='o0 = i0; Py_INCREF(o0);'),
            InstrSpec('return', pg('S', ''),
                       code='return i0;',
                       handles_own_decref=True),
            InstrSpec('py_call', pg('C[D]S@D', 'S'),
                       handles_own_decref=True,
                       code=ri(0, """
                           PyObject *py_args = PyTuple_New(n_i1);
                           if (py_args == NULL) goto error;
                           int i;
                           for (i = 0; i < n_i1; i++) {
                             PyObject *arg = i1[i];
                             PyTuple_SET_ITEM(py_args, i, arg);
                             i1[i] = NULL;
                           }
                           o0 = PyObject_CallObject(i0, py_args);
                           Py_DECREF(py_args);
                           """))
        ]

        binops = [
            ('add', 'PyNumber_Add'),
            ('sub', 'PyNumber_Subtract'),
            ('mul', 'PyNumber_Multiply'),
            ('div', 'PyNumber_TrueDivide'),
            ('floordiv', 'PyNumber_FloorDivide')
        ]

        for (name, op) in binops:
            instrs.append(instr_funcall_2args(name, pg('SS', 'S'), op))
        instrs.append(InstrSpec('pow', pg('SS', 'S'),
                                code='o0 = PyNumber_Power(i0, i1, Py_None);'))
        instrs.append(InstrSpec('ipow', pg('SC[D]', 'S'),
                                code='o0 = PyNumber_Power(i0, i1, Py_None);'))
        for (name, op) in [('neg', 'PyNumber_Negative'),
                           ('invert', 'PyNumber_Invert'),
                           ('abs', 'PyNumber_Absolute')]:
            instrs.append(instr_unary(name, pg('S', 'S'), '%s(i0)'%op))
        self.instr_descs = instrs
        self._set_opcodes()
        # Always use ipow
        self.ipow_range = True
        # We don't yet support call_c for Python-object interpreters
        # (the default implementation doesn't work, because of
        # object vs. PyObject* confusion)
        self.implement_call_c = False



# -*- coding: utf-8 -*-
"""
Set Up Logging

Logging can be customized using the ``SAGE_BOOTSTRAP`` environment
variable. It is a comma-separated list of ``key:value`` pairs. They
are not case sensitive. Valid pairs are:

* ``log:[level]``, where ``[level]`` is one of 

    * ``debug``
    * ``info``
    * ``warning``
    * ``critical``
    * ``error``

* ``interactive:true`` or ``interactive:false``, to override isatty detection.
"""


# ****************************************************************************
#       Copyright (C) 2015 Volker Braun <vbraun.name@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************


import sys
import os


LOG_LEVELS = (
    'debug',
    'info',
    'warning',
    'critical',
    'error'
)


class Configuration(object):

    _initialized = False

    log = 'info'

    interactive = os.isatty(sys.stdout.fileno())
    
    def __init__(self):
        if not Configuration._initialized:
            Configuration._init_from_environ()
        if self.log not in LOG_LEVELS:
            raise ValueError('invalid log level: {0}'.format(self.log))
        assert isinstance(self.interactive, bool)

    @classmethod
    def _init_from_environ(cls):
        env = os.environ.get('SAGE_BOOTSTRAP', '').lower()
        for pair in env.split(','):
            if not pair.strip():
                continue
            key, value = pair.split(':', 1)
            key = key.strip()
            value = value.strip()
            if key == 'log':
                cls.log = value
            elif key == 'interactive':
                if value == 'true':
                    cls.interactive = True
                elif value == 'false':
                    cls.interactive = False
                else:
                    raise ValueError('interactive value must be "true" or "false", got "{0}"'
                                     .format(value))
            else:
                raise ValueError('unknown key: "{0}"'.format(key))
        cls._initialized = True

    def __repr__(self):
        return '\n'.join([
            'Configuration:',
            '  * log = {0}'.format(self.log),
            '  * interactive = {0}'.format(self.interactive)
        ])



######################################################################
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
######################################################################

from __future__ import print_function

import platform
import os
import sys
import shutil
import glob
import subprocess


# this dictionary will hold all configuration information
conf = dict()

# Here is a list of keys and (some of) their possible values. First,
# strings:
#
# system:  Linux, SunOS, Darwin, FreeBSD, CYGWIN
# release: (Version number of the operating system, output of uname -r)
# machine: i386, x86_64,   # Linux, Darwin, *BSD
#          sun4u, i86pc    # SunOS
# processor: i386, x86_64, powerpc, sparc
# bits:    32bit, 64bit
# fortran: g95, gfortran
# ld:      GNU, Solaris, Darwin

# The following pre-defined boolean values are determined from the
# strings. The keys are distinguished by a question mark at the
# end. If possible use these keys as they guard against typos.
#
# Linux?, Solaris?, Darwin?, FreeBSD?, CYGWIN?  # system
# OS_X_Lion?                                    # release
# Intel?, PPC?, SPARC? ARM?                     # processor
# 64bit?, 32bit?                                # bits
# fortran_g95?, fortran_GNU?                    # fortran
# linker_GNU?, linker_Solaris?, linker_Darwin?  # ld


######################################################################
### Functions
######################################################################

def try_run(command, ignore=False):
    """
    Try to execute ``command`` and return its output as string.

    Return ``None`` if command not found. Return ``None`` if execution
    fails and ``ignore=False`` (default). Otherwise, return output of
    ``command`` as string. Here, output always means the concatenation
    of stdout and stderr.
    """
    f = subprocess.Popen(command, shell=True,
                         stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    result = f.communicate()
    rc = f.wait()
    if (not ignore) and (rc!=0):
        return None
    # concatenate stdout and stderr
    return (result[0].strip() + result[1].strip()).decode('utf8')


def cp(source_pattern, destination):
    """
    Portable implementation of "cp -p"
    """
    for filename in glob.iglob(source_pattern):
        print('Copying', filename, 'to', destination)
        shutil.copy2(filename, destination)


def ln(source, destination):
    """
    Portable implementation of "ln -sf"
    """
    if os.path.exists(destination):
        os.remove(destination)
    print('Linking', source, 'to', destination)
    os.symlink(source, destination)


def is_executable_file(fpath):
    """
    Check that `fpath` is a file and executable (ie not a directory with +x set)
    """
    return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

def which(program):
    """
    Portable implementation of "which"
    """
    fpath, fname = os.path.split(program)
    if fpath != '':
        return program
    for path in os.environ['PATH'].split(os.pathsep):
        f = os.path.join(path, program)
        if is_executable_file(f):
            return f
    raise ValueError('The program '+program+' is not in the $PATH.')

class edit_in_place(object):
    """
    Edit a file in-place by search/replacing regular expressions.
    """
    def __init__(self, filename):
        self.filename = filename
        f = open(filename, 'r')
        self.data = f.read()
        f.close()

    def replace(self, find_regex, subs, count=0):
        self.data = glob.re.sub(find_regex, subs, self.data, count)
        return self

    def close(self):
        f = open(self.filename, 'w')
        f.write(self.data)
        f.close()


######################################################################
### uname
######################################################################

try:
    conf['system'] = os.environ['UNAME']
except KeyError:
    conf['system'] = platform.system()

try:
    conf['release'] = subprocess.check_output(['uname', '-r']).decode('utf8').strip()
except subprocess.CalledProcessError:
    conf['release'] = ""

conf['Linux?'  ] = (conf['system'] == 'Linux')
conf['Solaris?'] = (conf['system'] == 'SunOS')
conf['Darwin?' ] = (conf['system'] == 'Darwin')
conf['OS_X_Lion?'] = (conf['Darwin?'] and conf['release'].startswith('11.'))
conf['FreeBSD?'] = (conf['system'] == 'FreeBSD')
conf['CYGWIN?' ] = (conf['system'] == 'CYGWIN')

conf['machine'] = platform.machine()

conf['processor'] = platform.processor()

conf['ARM?'] = (platform.machine().startswith('arm'))
conf['Intel?'] = (platform.machine() in ('i386', 'i486', 'i586', 'i686', 'x86_64',
                                         'AMD64', 'i86pc'))
conf['IA64?']   = ((platform.processor() == 'ia64')
                   or (platform.machine() == 'ia64'))
conf['PPC?']   = (platform.processor() == 'powerpc')
conf['SPARC?'] = (platform.processor() == 'sparc')

conf['generic_binary?'] = (os.environ.get('SAGE_FAT_BINARY', 'no') == 'yes')

conf['user'] = os.environ.get('ATLAS_CONFIGURE', '')

######################################################################
### bit width
######################################################################

conf['bits'] = platform.architecture()[0]

conf['64bit?'] = (conf['bits'] == '64bit')
conf['32bit?'] = (conf['bits'] != '64bit')


######################################################################
### fortran compiler
######################################################################

fortran_version = try_run('$FC --version')
if fortran_version is None:
    print('Cannot execute fortran compiler ($FC)!')
    sys.exit(3)
if 'G95' in fortran_version:
    conf['fortran'] = 'g95'
elif 'GNU Fortran' in fortran_version:
    conf['fortran'] = 'gfortran'
else:
    print('Unknown fortran compiler version: '+fortran_version)
    conf['fortran'] = None


conf['fortran_g95?'] = (conf['fortran'] == 'g95')
conf['fortran_GNU?'] = (conf['fortran'] == 'gfortran')


if conf['fortran_g95?']:
    g95_dir = glob.glob(os.environ['SAGE_LOCAL']+'/lib/gcc-lib/*/*')
    assert len(g95_dir)==1, 'Could not find G95 dir: '+str(g95_dir)
    conf['fortran_g95_dir'] = g95_dir[0]

######################################################################
### linker
######################################################################

if conf['Solaris?']:
    # Note: Solaris linker does not accept --version
    # ignore error code as 'ld -V' likes to error on some Solaris versions
    ld_version = try_run('ld  -V', ignore=True)
else:
    ld_version = try_run('ld  -v')

if ld_version is None:
    print('Cannot execute ld!')
    sys.exit(3)
if 'GNU' in ld_version:
    conf['ld'] = 'GNU'
elif 'Solaris' in ld_version:
    conf['ld'] = 'Solaris'
elif 'Apple' in ld_version:
    conf['ld'] = 'Darwin'
else:
    print('Unknown linker: '+ld_version)
    conf['ld'] = None

conf['linker_GNU?'] = (conf['ld'] == 'GNU')
conf['linker_Solaris?'] = (conf['ld'] == 'Solaris')
conf['linker_Darwin?'] = (conf['ld'] == 'Darwin')


if conf['Solaris?'] and conf['linker_GNU?']:
    print("WARNING: You are using the GNU linker from 'binutils'")
    print("Generally it is considered better to use the Sun linker")
    print("but Sage has been built on Solaris using the GNU linker")
    print("although that was a very old version of Sage, which")
    print("never passes all the Sage test-suite.")



######################################################################
### paths, files, and environment variables
######################################################################

conf['SPKG_DIR'] = os.getcwd()
conf['SAGE_LOCAL'] = os.environ['SAGE_LOCAL']


######################################################################
### The end: print configuration
######################################################################

print("Configuration:")
for k in sorted(conf, key=str.lower):
    print('    {}: {}'.format(k, conf[k]))



########################################################################
##
## Parallel gcc execution
##
## This code is responsible for making distutils dispatch the calls to
## build_ext in parallel. Since distutils doesn't seem to do this by
## default, we create our own extension builder and override the
## appropriate methods.  Unfortunately, in distutils, the logic of
## deciding whether an extension needs to be recompiled and actually
## making the call to gcc to recompile the extension are in the same
## function. As a result, we can't just override one function and have
## everything magically work. Instead, we split this work between two
## functions. This works fine for our application, but it means that
## we can't use this modification to make the other parts of Sage that
## build with distutils call gcc in parallel.
##
########################################################################

import os
import sys
import time

keep_going = False

def run_command(cmd):
    """
    INPUT:

    - ``cmd`` -- a string; a command to run

    OUTPUT: prints ``cmd`` to the console and then runs
    ``os.system(cmd)``.
    """
    print(cmd)
    sys.stdout.flush()
    return os.system(cmd)

def apply_func_progress(p):
    """
    Given a triple p consisting of a function, value and a string,
    output the string and apply the function to the value.

    The string could for example be some progress indicator.

    This exists solely because we can't pickle an anonymous function
    in execute_list_of_commands_in_parallel below.
    """
    sys.stdout.write(p[2])
    sys.stdout.flush()
    return p[0](p[1])

def execute_list_of_commands_in_parallel(command_list, nthreads):
    """
    Execute the given list of commands, possibly in parallel, using
    ``nthreads`` threads.  Terminates ``setup.py`` with an exit code
    of 1 if an error occurs in any subcommand.

    INPUT:

    - ``command_list`` -- a list of commands, each given as a pair of
       the form ``[function, argument]`` of a function to call and its
       argument

    - ``nthreads`` -- integer; number of threads to use

    WARNING: commands are run roughly in order, but of course successive
    commands may be run at the same time.
    """
    # Add progress indicator strings to the command_list
    N = len(command_list)
    progress_fmt = "[{:%i}/{}] " % len(str(N))
    for i in range(N):
        progress = progress_fmt.format(i+1, N)
        command_list[i] = command_list[i] + (progress,)

    from multiprocessing import Pool
    # map_async handles KeyboardInterrupt correctly if an argument is
    # given to get().  Plain map() and apply_async() do not work
    # correctly, see Trac #16113.
    pool = Pool(nthreads)
    result = pool.map_async(apply_func_progress, command_list, 1).get(99999)
    pool.close()
    pool.join()
    process_command_results(result)

def process_command_results(result_values):
    error = None
    for r in result_values:
        if r:
            print("Error running command, failed with status %s."%r)
            if not keep_going:
                sys.exit(1)
            error = r
    if error:
        sys.exit(1)

def execute_list_of_commands(command_list):
    """
    INPUT:

    - ``command_list`` -- a list of strings or pairs

    OUTPUT:

    For each entry in command_list, we attempt to run the command.
    If it is a string, we call ``os.system()``. If it is a pair [f, v],
    we call f(v).

    If the environment variable :envvar:`SAGE_NUM_THREADS` is set, use
    that many threads.
    """
    t = time.time()
    # Determine the number of threads from the environment variable
    # SAGE_NUM_THREADS, which is set automatically by sage-env
    try:
        nthreads = int(os.environ['SAGE_NUM_THREADS'])
    except KeyError:
        nthreads = 1

    # normalize the command_list to handle strings correctly
    command_list = [ [run_command, x] if isinstance(x, str) else x for x in command_list ]

    # No need for more threads than there are commands, but at least one
    nthreads = min(len(command_list), nthreads)
    nthreads = max(1, nthreads)

    def plural(n,noun):
        if n == 1:
            return "1 %s"%noun
        return "%i %ss"%(n,noun)

    print("Executing %s (using %s)"%(plural(len(command_list),"command"), plural(nthreads,"thread")))
    execute_list_of_commands_in_parallel(command_list, nthreads)
    print("Time to execute %s: %.2f seconds."%(plural(len(command_list),"command"), time.time() - t))



#########################################################
### Library order
#########################################################

# This list defines the *order* of linking libraries. A library should
# be put *before* any library it links to. Cython allows
# defining libraries using "# distutils: libraries = LIB". However, if
# there are multiple libraries, the order is undefined so we need to
# manually reorder the libraries according to this list. The order is
# important in particular for Cygwin. Any libraries which are not
# listed here will be added at the end of the list (without changing
# their relative order).
from sage.env import cython_aliases
aliases = cython_aliases()

arb_dylib_name = aliases["ARB_LIBRARY"]
library_order_list = aliases["SINGULAR_LIBRARIES"] + [
    "giac", "intl", "curl",
    "ec", "ecm"
] + aliases["LINBOX_LIBRARIES"] + aliases["FFLASFFPACK_LIBRARIES"] + aliases["GSL_LIBRARIES"] + [
    "pari", "flint", "ratpoints", "ecl", "glpk", "ppl",
    arb_dylib_name, "mpfi", "mpfr", "mpc", "ntl", "gmp", "gmpxx",
    "brial",
    "brial_groebner",
    "m4rie",
] + aliases["M4RI_LIBRARIES"] + [
    "zn_poly", "gap",
] + aliases["GDLIB_LIBRARIES"] + aliases["LIBPNG_LIBRARIES"] + [
    "m", "readline", "Lfunction" ,
] + aliases["CBLAS_LIBRARIES"] + aliases["ZLIB_LIBRARIES"]

# Make a dict with library:order pairs, where the order are negative
# integers sorted according to library_order_list. When sorting,
# unlisted libraries have order 0, so they appear after the libraries
# in library_order_list.
n = len(library_order_list)
library_order = {}
for i in range(n):
    lib = library_order_list[i]
    library_order[lib] = i-n



# -*- coding: utf-8 -*-
"""
Utility to manage lists of packages
"""

# ****************************************************************************
#       Copyright (C) 2016 Volker Braun <vbraun.name@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

import logging
from sage_bootstrap.package import Package

log = logging.getLogger()


class PackageClass(object):

    def __init__(self, *package_names_or_classes, **filters):
        self.names = []
        filenames = filters.pop('has_files', [])
        if filters:
            raise ValueError('filter not supported')

        def included_in_filter(pkg):
            return all(pkg.has_file(filename) for filename in filenames)
        for package_name_or_class in package_names_or_classes:
            if package_name_or_class == ':all:':
                self._init_all(predicate=included_in_filter)
            elif package_name_or_class == ':standard:':
                self._init_standard(predicate=included_in_filter)
            elif package_name_or_class == ':optional:':
                self._init_optional(predicate=included_in_filter)
            elif package_name_or_class == ':experimental:':
                self._init_experimental(predicate=included_in_filter)
            elif package_name_or_class == ':huge:':
                self._init_huge(predicate=included_in_filter)
            else:
                if ':' in package_name_or_class:
                    raise ValueError('a colon may only appear in designators of package types, '
                                     'which must be one of '
                                     ':all:, :standard:, :optional:, :experimental:, or :huge:, '
                                     'got {}'.format(package_name_or_class))
                self.names.append(package_name_or_class)

    def _init_all(self, predicate):
        self.names.extend(pkg.name for pkg in Package.all() if predicate(pkg))

    def _init_standard(self, predicate):
        self.names.extend(pkg.name for pkg in Package.all() if pkg.type == 'standard' and predicate(pkg))

    def _init_optional(self, predicate):
        self.names.extend(pkg.name for pkg in Package.all() if pkg.type == 'optional' and predicate(pkg))

    def _init_experimental(self, predicate):
        self.names.extend(pkg.name for pkg in Package.all() if pkg.type == 'experimental' and predicate(pkg))

    def _init_huge(self, predicate):
        self.names.extend(pkg.name for pkg in Package.all() if pkg.type == 'huge' and predicate(pkg))

    def apply(self, function, *args, **kwds):
        for package_name in self.names:
            function(package_name, *args, **kwds)



# pkgs/sage-conf/sage_conf.py.  Generated from sage_conf.py.in by configure.

VERSION = "9.4"

# The following must not be used during build to determine source or installation
# location of sagelib.  See comments in SAGE_ROOT/src/Makefile.in
# These variables come first so that other substituted variable values can refer
# to it.
SAGE_LOCAL = "/Users/mkoeppe/s/sage/sage-rebasing/worktree-dist/.tox/local-macos-10.15-nohomebrew-python3_xcode/local"
SAGE_ROOT = "/Users/mkoeppe/s/sage/sage-rebasing/worktree-dist"

MAXIMA = "/Users/mkoeppe/s/sage/sage-rebasing/worktree-dist/.tox/local-macos-10.15-nohomebrew-python3_xcode/local/bin/maxima"

# Delete this line if your ECL can load maxima without further prodding.
MAXIMA_FAS = "${prefix}/lib/ecl/maxima.fas".replace('${prefix}', SAGE_LOCAL)

# Delete this line if your ECL can load Kenzo without further prodding.
KENZO_FAS = "${prefix}/lib/ecl/kenzo.fas".replace('${prefix}', SAGE_LOCAL)

ARB_LIBRARY = "arb"

NTL_INCDIR = ""
NTL_LIBDIR = ""

# Path to the ecl-config script
ECL_CONFIG = "${prefix}/bin/ecl-config".replace('${prefix}', SAGE_LOCAL)

SAGE_NAUTY_BINS_PREFIX = ""

# Colon-separated list of pkg-config modules to search for cblas functionality.
# We hard-code it here as cblas because configure (build/pkgs/openblas/spkg-configure.m4)
# always provides cblas.pc, if necessary by creating a facade pc file for a system BLAS.
CBLAS_PC_MODULES = "cblas"

# for sage_setup.setenv
SAGE_ARCHFLAGS = ""
SAGE_PKG_CONFIG_PATH = "".replace('$SAGE_LOCAL', SAGE_LOCAL)

# Used in sage.repl.ipython_kernel.install
MATHJAX_DIR = SAGE_LOCAL + "/share/mathjax"
THREEJS_DIR = SAGE_LOCAL + "/share/threejs-sage"

# OpenMP flags, if available.
OPENMP_CFLAGS = ""
OPENMP_CXXFLAGS = ""

SAGE_SPKG_WHEELS = "/Users/mkoeppe/s/sage/sage-rebasing/worktree-dist/.tox/local-macos-10.15-nohomebrew-python3_xcode/local/var/lib/sage/wheels"

# Entry point 'sage-config'.  It does not depend on any packages.

def _main():
    from argparse import ArgumentParser
    from sys import exit, stdout
    parser = ArgumentParser()
    parser.add_argument('--version', help="show version", action="version",
                       version='%(prog)s ' + VERSION)
    parser.add_argument("VARIABLE", nargs='?', help="output the value of VARIABLE")
    args = parser.parse_args()
    d = globals()
    if args.VARIABLE:
        stdout.write('{}\n'.format(d[args.VARIABLE]))
    else:
        for k, v in d.items():
            if not k.startswith('_'):
                stdout.write('{}={}\n'.format(k, v))

if __name__ == "__main__":
    _main()



"""Miscellaneous utilities for running the docbuilder."""

import errno
import os


class WorkerDiedException(RuntimeError):
    """Raised if a worker process dies unexpected."""

    def __init__(self, message, original_exception=None):
        super(WorkerDiedException, self).__init__(message)
        self.original_exception = original_exception


def build_many(target, args, processes=None):
    """
    Map a list of arguments in ``args`` to a single-argument target function
    ``target`` in parallel using ``multiprocessing.cpu_count()`` (or
    ``processes`` if given) simultaneous processes.

    This is a simplified version of ``multiprocessing.Pool.map`` from the
    Python standard library which avoids a couple of its pitfalls.  In
    particular, it can abort (with a `RuntimeError`) without hanging if one of
    the worker processes unexpectedly dies.  It also has semantics equivalent
    to ``maxtasksperchild=1``; that is, one process is started per argument.
    As such, this is inefficient for processing large numbers of fast tasks,
    but appropriate for running longer tasks (such as doc builds) which may
    also require significant cleanup.

    It also avoids starting new processes from a pthread, which results in at
    least two known issues:

        * On versions of Cygwin prior to 3.0.0 there were bugs in mmap handling
          on threads (see https://trac.sagemath.org/ticket/27214#comment:25).

        * When PARI is built with multi-threading support, forking a Sage
          process from a thread leaves the main Pari interface instance broken
          (see https://trac.sagemath.org/ticket/26608#comment:38).

    In the future this may be replaced by a generalized version of the more
    robust parallel processing implementation from ``sage.doctest.forker``.

    EXAMPLES::

        sage: from sage_docbuild.utils import build_many
        sage: def target(N):
        ....:     import time
        ....:     time.sleep(float(0.1))
        ....:     print('Processed task %s' % N)
        ....:
        sage: _ = build_many(target, range(8), processes=8)
        Processed task ...
        Processed task ...
        Processed task ...
        Processed task ...
        Processed task ...
        Processed task ...
        Processed task ...
        Processed task ...

    Unlike the first version of `build_many` which was only intended to get
    around the Cygwin bug, this version can also return a result, and thus can
    be used as a replacement for `multiprocessing.Pool.map` (i.e. it still
    blocks until the result is ready)::

        sage: def square(N):
        ....:     return N * N
        sage: build_many(square, range(100))
        [0, 1, 4, 9, ..., 9604, 9801]

    If the target function raises an exception in any of the workers,
    `build_many` raises that exception and all other results are discarded.
    Any in-progress tasks may still be allowed to complete gracefully before
    the exception is raised::

        sage: def target(N):
        ....:     import time, os, signal
        ....:     if N == 4:
        ....:         # Task 4 is a poison pill
        ....:         1 / 0
        ....:     else:
        ....:         time.sleep(0.5)
        ....:         print('Processed task %s' % N)
        ....:

    Note: In practice this test might still show output from the other worker
    processes before the poison-pill is executed.  It may also display the
    traceback from the failing process on stderr.  However, due to how the
    doctest runner works, the doctest will only expect the final exception::

        sage: build_many(target, range(8), processes=8)
        Traceback (most recent call last):
        ...
        ZeroDivisionError: rational division by zero

    Similarly, if one of the worker processes dies unexpectedly otherwise exits
    non-zero (e.g. killed by a signal) any in-progress tasks will be completed
    gracefully, but then a `RuntimeError` is raised and pending tasks are not
    started::

        sage: def target(N):
        ....:     import time, os, signal
        ....:     if N == 4:
        ....:         # Task 4 is a poison pill
        ....:         os.kill(os.getpid(), signal.SIGKILL)
        ....:     else:
        ....:         time.sleep(0.5)
        ....:         print('Processed task %s' % N)
        ....:
        sage: build_many(target, range(8), processes=8)
        Traceback (most recent call last):
        ...
        WorkerDiedException: worker for 4 died with non-zero exit code -9
    """
    from multiprocessing import Process, Queue, cpu_count, set_start_method
    # With OS X, Python 3.8 defaults to use 'spawn' instead of 'fork'
    # in multiprocessing, and Sage docbuilding doesn't work with
    # 'spawn'. See trac #27754.
    if os.uname().sysname == 'Darwin':
        set_start_method('fork', force=True)
    from queue import Empty

    if processes is None:
        processes = cpu_count()

    workers = [None] * processes
    tasks = enumerate(args)
    results = []
    result_queue = Queue()

    # Utility functions #
    def run_worker(target, queue, idx, task):
        try:
            result = target(task)
        except BaseException as exc:
            queue.put((None, exc))
        else:
            queue.put((idx, result))

    def bring_out_yer_dead(w, task, exitcode):
        """
        Handle a dead / completed worker.  Raises WorkerDiedException if it
        returned with a non-zero exit code.
        """

        if w is None or exitcode is None:
            # I'm not dead yet! (or I haven't even been born yet)
            return (w, task)

        # Hack: If we wait()ed on this worker manually we have to tell it
        # it's dead:
        if w._popen.returncode is None:
            w._popen.returncode = exitcode

        if exitcode != 0:
            raise WorkerDiedException(
                "worker for {} died with non-zero exit code "
                "{}".format(task[1], w.exitcode))

        # Get result from the queue; depending on ordering this may not be
        # *the* result for this worker, but for each completed worker there
        # should be *a* result so let's get it
        try:
            result = result_queue.get_nowait()
        except Empty:
            # Generally shouldn't happen but could in case of a race condition;
            # don't worry we'll collect any remaining results at the end.
            pass

        if result[0] is None:
            # Indicates that an exception occurred in the target function
            raise WorkerDiedException('', original_exception=result[1])
        else:
            results.append(result)

        # Helps multiprocessing with some internal bookkeeping
        w.join()

        return None

    def wait_for_one():
        """Wait for a single process and return its pid and exit code."""
        try:
            pid, sts = os.wait()
        except OSError as exc:
            # No more processes to wait on if ECHILD
            if exc.errno != errno.ECHILD:
                raise
            else:
                return None, None

        if os.WIFSIGNALED(sts):
            exitcode = -os.WTERMSIG(sts)
        else:
            exitcode = os.WEXITSTATUS(sts)

        return pid, exitcode

    def reap_workers(waited_pid=None, waited_exitcode=None):
        """
        This is the main worker handling loop.

        Checks if workers have completed their tasks and spawns new workers if
        there are more tasks on the queue.  Returns `False` if there is more
        work to be done or `True` if the work is complete.

        Raises a ``WorkerDiedException`` if a worker exits unexpectedly.
        """

        all_done = True

        for idx, w in enumerate(workers):
            if w is not None:
                w, task = w
                if w.pid == waited_pid:
                    exitcode = waited_exitcode
                else:
                    exitcode = w.exitcode

                w = bring_out_yer_dead(w, task, exitcode)

            # Worker w is dead/not started, so start a new worker
            # in its place with the next document from the queue
            if w is None:
                try:
                    task = next(tasks)
                except StopIteration:
                    pass
                else:
                    w = Process(target=run_worker,
                                args=((target, result_queue) + task))
                    w.start()
                    # Pair the new worker with the task it's performing (mostly
                    # for debugging purposes)
                    w = (w, task)

            workers[idx] = w

            if w is not None:
                all_done = False

        # If all workers are dead and there are no more items to
        # process in the queue then we are done
        return all_done

    # Main loop #

    waited_pid = None
    # Set along with waited_exitcode by calls to wait_for_one()

    waited_exitcode = None
    worker_exc = None  # Set to a WorkerDiedException if one occurs

    try:
        while True:
            # Check the status of each worker and break out of the loop if
            # all work is done.
            # We'll check each worker process against the returned
            # pid back at the top of the `while True` loop.  We also
            # check any other processes that may have exited in the
            # meantime
            try:
                if reap_workers(waited_pid, waited_exitcode):
                    break
            except WorkerDiedException as exc:
                worker_exc = exc
                break

            waited_pid, waited_exitcode = wait_for_one()
    finally:
        try:
            remaining_workers = [w for w in workers if w is not None]
            for w, _ in remaining_workers:
                # Give any remaining workers a chance to shut down gracefully
                try:
                    w.terminate()
                except OSError as exc:
                    if exc.errno != errno.ESRCH:
                        # Otherwise it was already dead so this was expected
                        raise
            for w, _ in remaining_workers:
                w.join()
        finally:
            if worker_exc is not None:
                # Re-raise the RuntimeError from bring_out_yer_dead set if a
                # worker died unexpectedly, or the original exception if it's
                # wrapping one
                if worker_exc.original_exception:
                    raise worker_exc.original_exception
                else:
                    raise worker_exc

    # All workers should be shut down by now and should have completed without
    # error.  No new items will be added to the result queue, so we can get all
    # the remaining results, if any.
    while True:
        try:
            results.append(result_queue.get_nowait())
        except Empty:
            break

    # Return the results sorted according to their original task order
    return [r[1] for r in sorted(results, key=lambda r: r[0])]



"""
Optional extensions

An "optional extension" is a Cython extension which is always
cythonized (i.e. converted to a .c or .cpp file), but which is only
compiled depending on some condition. Typically, this condition is a
package which must be installed.
"""

# ****************************************************************************
#       Copyright (C) 2015 Jeroen Demeyer <jdemeyer@cage.ugent.be>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from setuptools.extension import Extension
from sage.misc.package import list_packages

all_packages = list_packages(local=True)


class CythonizeExtension(Extension):
    """
    A class for extensions which are only cythonized, but not built.

    The file ``src/setup.py`` contains some logic to check the
    ``skip_build`` attribute of extensions.

    EXAMPLES::

        sage: from sage_setup.optional_extension import CythonizeExtension
        sage: ext = CythonizeExtension("foo", ["foo.c"])
        sage: ext.skip_build
        True
    """
    skip_build = True


def is_package_installed_and_updated(pkg):
    from sage.misc.package import is_package_installed
    try:
        pkginfo = all_packages[pkg]
    except KeyError:
        # Might be an installed old-style package
        condition = is_package_installed(pkg)
    else:
        condition = (pkginfo["installed_version"] == pkginfo["remote_version"])
    return condition


def OptionalExtension(*args, **kwds):
    """
    If some condition (see INPUT) is satisfied, return an ``Extension``.
    Otherwise, return a ``CythonizeExtension``.

    Typically, the condition is some optional package or something
    depending on the operating system.

    INPUT:

    - ``condition`` -- (boolean) the actual condition

    - ``package`` -- (string) the condition is that this package is
      installed and up-to-date (only used if ``condition`` is not given)

    EXAMPLES::

        sage: from sage_setup.optional_extension import OptionalExtension
        sage: ext = OptionalExtension("foo", ["foo.c"], condition=False)
        sage: print(ext.__class__.__name__)
        CythonizeExtension
        sage: ext = OptionalExtension("foo", ["foo.c"], condition=True)
        sage: print(ext.__class__.__name__)
        Extension
        sage: ext = OptionalExtension("foo", ["foo.c"], package="no_such_package")
        sage: print(ext.__class__.__name__)
        CythonizeExtension
        sage: ext = OptionalExtension("foo", ["foo.c"], package="gap")
        sage: print(ext.__class__.__name__)
        Extension
    """
    try:
        condition = kwds.pop("condition")
    except KeyError:
        pkg = kwds.pop("package")
        condition = is_package_installed_and_updated(pkg)
    if condition:
        return Extension(*args, **kwds)
    else:
        return CythonizeExtension(*args, **kwds)



# -*- coding: utf-8 -*-
"""
Test the is_url utility
"""

# ****************************************************************************
#       Copyright (C) 2016 Volker Braun <vbraun.name@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************


import unittest

from sage_bootstrap.util import is_url


class IsUrlTestCase(unittest.TestCase):

    def test_http(self):
        self.assertTrue(is_url('http://foo.bar/baz'))

    def test_https(self):
        self.assertTrue(is_url('https://foo.bar/baz'))

    def test_ftp(self):
        self.assertTrue(is_url('ftp://foo.bar/baz'))

    def test_nospace(self):
        self.assertFalse(is_url('http://foo. bar/baz'))

    def test_single_line(self):
        self.assertFalse(is_url('http://foo.bar/baz\nhttp://foo.bar/baz'))



from setuptools.command.egg_info import egg_info as setuptools_egg_info

class sage_egg_info(setuptools_egg_info):

    def finalize_options(self):
        from sage.env import SAGE_CONF_WHEEL_URL
        if SAGE_CONF_WHEEL_URL:
            self.distribution.install_requires.append(
                f'sage_conf @ {SAGE_CONF_WHEEL_URL}')
        setuptools_egg_info.finalize_options(self)



import os
import errno

# Import setuptools before importing distutils, so that setuptools
# can replace distutils by its own vendored copy.
import setuptools

from distutils import log
from setuptools.command.build_ext import build_ext
from distutils.dep_util import newer_group
from distutils.errors import DistutilsSetupError
from sage_setup.run_parallel import execute_list_of_commands

class sage_build_ext(build_ext):
    def finalize_options(self):
        build_ext.finalize_options(self)
        self.check_flags()

    def run(self):
        # Always run the Cythonize command before building extensions
        self.run_command('build_cython')
        build_ext.run(self)

    def check_flags(self):
        """
        Sanity check the compiler flags used to build the extensions
        """
        forbidden = None
        if os.environ.get("SAGE_FAT_BINARY") == "yes":
            # When building with SAGE_FAT_BINARY=yes, we should not
            # enable CPU features which do not exist on every CPU.
            # Such flags usually come from other libraries adding the
            # flags to the pkgconfig configuration.  So if you hit these
            # errors, the problem is most likely with some external
            # library and not with Sage.
            import re
            forbidden = re.compile(r"-march=|-mpcu=|-msse3|-msse4|-mpopcnt|-mavx")

        if forbidden is not None:
            errors = 0
            for ext in self.extensions:
                flags = ext.extra_compile_args
                for flag in flags:
                    if forbidden.match(flag):
                        log.error("%s uses forbidden flag '%s'", ext.name, flag)
                        errors += 1
            if errors:
                raise RuntimeError("forbidden flags used")

    def build_extensions(self):

        from distutils.debug import DEBUG

        if DEBUG:
            print("self.compiler.compiler:")
            print(self.compiler.compiler)
            print("self.compiler.compiler_cxx:")
            print(self.compiler.compiler_cxx) # currently not used
            print("self.compiler.compiler_so:")
            print(self.compiler.compiler_so)
            print("self.compiler.linker_so:")
            print(self.compiler.linker_so)
            # There are further interesting variables...

        if DEBUG:
            print("self.compiler.linker_so (after fixing library dirs):")
            print(self.compiler.linker_so)


        # First, sanity-check the 'extensions' list
        self.check_extensions_list(self.extensions)

        import time
        t = time.time()

        compile_commands = []
        for ext in self.extensions:
            need_to_compile, p = self.prepare_extension(ext)
            if need_to_compile:
                compile_commands.append((self.build_extension, p))

        execute_list_of_commands(compile_commands)

        print("Total time spent compiling C/C++ extensions: %.2f seconds." % (time.time() - t))

    def prepare_extension(self, ext):
        sources = ext.sources
        if sources is None or not isinstance(sources, (list, tuple)):
            raise DistutilsSetupError(("in 'ext_modules' option (extension '%s'), " +
                   "'sources' must be present and must be " +
                   "a list of source filenames") % ext.name)
        sources = list(sources)

        fullname = self.get_ext_fullname(ext.name)
        if self.inplace:
            # ignore build-lib -- put the compiled extension into
            # the source tree along with pure Python modules

            modpath = fullname.split('.')
            package = '.'.join(modpath[0:-1])
            base = modpath[-1]

            build_py = self.get_finalized_command('build_py')
            package_dir = build_py.get_package_dir(package)
            ext_filename = os.path.join(package_dir,
                                        self.get_ext_filename(base))
            relative_ext_filename = self.get_ext_filename(base)
        else:
            ext_filename = os.path.join(self.build_lib,
                                        self.get_ext_filename(fullname))
            relative_ext_filename = self.get_ext_filename(fullname)

        # while dispatching the calls to gcc in parallel, we sometimes
        # hit a race condition where two separate build_ext objects
        # try to create a given directory at the same time; whoever
        # loses the race then seems to throw an error, saying that
        # the directory already exists. so, instead of fighting to
        # fix the race condition, we simply make sure the entire
        # directory tree exists now, while we're processing the
        # extensions in serial.
        relative_ext_dir = os.path.split(relative_ext_filename)[0]
        prefixes = ['', self.build_lib, self.build_temp]
        for prefix in prefixes:
            path = os.path.join(prefix, relative_ext_dir)
            try:
                os.makedirs(path)
            except OSError as e:
                assert e.errno==errno.EEXIST, 'Cannot create %s.' % path
        depends = sources + ext.depends
        if not (self.force or newer_group(depends, ext_filename, 'newer')):
            log.debug("skipping '%s' extension (up-to-date)", ext.name)
            need_to_compile = False
        else:
            log.info("building '%s' extension", ext.name)
            need_to_compile = True

        return need_to_compile, (sources, ext, ext_filename)

    def build_extension(self, p):

        sources, ext, ext_filename = p

        # First, scan the sources for SWIG definition files (.i), run
        # SWIG on 'em to create .c files, and modify the sources list
        # accordingly.
        sources = self.swig_sources(sources, ext)

        # Next, compile the source code to object files.

        # XXX not honouring 'define_macros' or 'undef_macros' -- the
        # CCompiler API needs to change to accommodate this, and I
        # want to do one thing at a time!

        # Two possible sources for extra compiler arguments:
        #   - 'extra_compile_args' in Extension object
        #   - CFLAGS environment variable (not particularly
        #     elegant, but people seem to expect it and I
        #     guess it's useful)
        # The environment variable should take precedence, and
        # any sensible compiler will give precedence to later
        # command line args.  Hence we combine them in order:
        extra_args = ext.extra_compile_args or []

        macros = ext.define_macros[:]
        for undef in ext.undef_macros:
            macros.append((undef,))

        objects = self.compiler.compile(sources,
                                        output_dir=self.build_temp,
                                        macros=macros,
                                        include_dirs=ext.include_dirs,
                                        debug=self.debug,
                                        extra_postargs=extra_args,
                                        depends=ext.depends)

        # XXX -- this is a Vile HACK!
        #
        # The setup.py script for Python on Unix needs to be able to
        # get this list so it can perform all the clean up needed to
        # avoid keeping object files around when cleaning out a failed
        # build of an extension module.  Since Distutils does not
        # track dependencies, we have to get rid of intermediates to
        # ensure all the intermediates will be properly re-built.
        #
        self._built_objects = objects[:]

        # Now link the object files together into a "shared object" --
        # of course, first we have to figure out all the other things
        # that go into the mix.
        if ext.extra_objects:
            objects.extend(ext.extra_objects)
        extra_args = ext.extra_link_args or []

        # Detect target language, if not provided
        language = ext.language or self.compiler.detect_language(sources)

        self.compiler.link_shared_object(
            objects, ext_filename,
            libraries=self.get_libraries(ext),
            library_dirs=ext.library_dirs,
            runtime_library_dirs=ext.runtime_library_dirs,
            extra_postargs=extra_args,
            export_symbols=self.get_export_symbols(ext),
            debug=self.debug,
            build_temp=self.build_temp,
            target_lang=language)



######################################################################
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
######################################################################

from __future__ import print_function

######################################################################
### Import stuff
######################################################################

import os, sys, shutil, time, glob
from configuration import conf, cp, ln, which, try_run, edit_in_place
from enums import ATLAS_OSTYPE, ATLAS_MACHTYPE, ATLAS_ISAEXT, make_check_enums


######################################################################
### The following variables may need to be updated if you
### update ATLAS or LAPACK to a newer version
######################################################################

PATCH_DIR = os.path.join(conf['SPKG_DIR'], 'patches')

# the current lapack source tarball
LAPACK_TARFILE = os.path.join(conf['SPKG_DIR'], 'src', 'lapack-3.5.0.tar')

# temporary directory to build everything in
BUILD_DIR = os.path.join(conf['SPKG_DIR'], 'src', 'ATLAS-build')

# the shared library autotools stub project
BUILD_LIB_DIR = os.path.join(conf['SPKG_DIR'], 'src', 'ATLAS-lib')

# we need to disable parallel builds
os.environ['MAKE'] += ' -j1'
MAKE = os.environ['MAKE']

# SAGE_LOCAL
SAGE_LOCAL = os.environ['SAGE_LOCAL']

######################################################################
### Some auxiliary functions to facilitate IO and error checking
######################################################################

# Run shell command "command", but flush stdout and stderr before doing
# this. Also echo commands which are executed.
def system_with_flush(command):
    print('Running', command)
    sys.stdout.flush()
    sys.stderr.flush()
    import subprocess
    return subprocess.call(command, shell=True)

def assert_success(rc, good=None, bad=None):
    if rc == 0:
        if good is not None:
            print(good)
        return
    print('-'*60)
    import traceback
    traceback.print_stack(file=sys.stdout)
    print('-'*60)
    if bad is not None:
        print('Error:', bad)
    sys.exit(rc)


######################################################################
### function to save ATLAS' configuration to .pc files
######################################################################

def write_pc_file(libs,target):
    pkgconfigdir=os.path.join(SAGE_LOCAL, 'lib', 'pkgconfig')
    if not os.path.isdir(pkgconfigdir):
        os.makedirs(pkgconfigdir)
    libflags='-l%s'%(' -l'.join(libs))
    pcfile ="""SAGE_LOCAL=%s
prefix=${SAGE_LOCAL}
libdir=${prefix}/lib
includedir=${prefix}/include
Name: %s
Version: 1.0
Description: %s for sage, set up by the ATLAS spkg.
Libs: -L${libdir} %s
"""%(SAGE_LOCAL, target, target, libflags)
    open(os.path.join(SAGE_LOCAL, 'lib/pkgconfig/%s.pc'%target), 'w').write(pcfile)


######################################################################
### Skip building ATLAS on specific systems
######################################################################

if conf['Darwin?'] and 'SAGE_ATLAS_ARCH' not in os.environ:
    print('Skipping build of ATLAS on OS X, using system library instead.')
    print('You can try building your own ATLAS by setting SAGE_ATLAS_ARCH')
    print('to something sensible although that is not officially supported.')
    if conf['PPC?']:   # OSX 10.4 PPC linker needs help to find the accelerate blas
        veclib_dir = '/System/Library/Frameworks/Accelerate.framework/' + \
            'Versions/A/Frameworks/vecLib.framework/Versions/A'
        for lib in [ 'libBLAS.dylib', 'libLAPACK.dylib']:
            ln(os.path.join(veclib_dir, lib),
               os.path.join(conf['SAGE_LOCAL'], 'lib', lib))
    write_pc_file(['blas'], 'cblas')
    write_pc_file(['blas'], 'blas')
    write_pc_file(['lapack'], 'lapack')
    sys.exit(0)


######################################################################
### Use SAGE_ATLAS_LIB='directory' if provided
######################################################################

# On Cygwin, we used to require that the system-wide lapack packages were
# installed (this included the lapack-devel and lapack packages).
# These packages indeed include BLAS and LAPACK and are enough to build
# the rest of Sage, whereas building ATLAS was problematic.
# For the record, the corresponding files to symlink from the system-wide
# packages are: '/usr/lib/libblas.dll.a' and  '/usr/lib/liblapack.dll.a'.

if 'SAGE_ATLAS_LIB' in os.environ:
    ATLAS_LIB = os.environ['SAGE_ATLAS_LIB']

    prefix = 'lib'

    libraries_sets = [['lapack', 'cblas', 'f77blas', 'atlas'], ['lapack', 'blas']]
    libraries_optional = ['ptcblas', 'ptf77blas']

    def is_atlas_lib_path(path, libs):
        if path is None:
            return False
        if not os.path.isdir(path):
            return False
        filenames = os.listdir(path)
        for lib in libs:
            if not any(fname.startswith(prefix+lib) for fname in filenames):
                print('Cannot find '+prefix+lib+'.* in '+path)
                break
        else:
            return True
        return False

    paths = [ ATLAS_LIB, os.path.join(ATLAS_LIB, 'lib64'), os.path.join(ATLAS_LIB, 'lib') ]
    ATLAS_LIB = None
    libraries = []
    for libs in libraries_sets:
        for path in paths:
            if is_atlas_lib_path(path, libs):
                ATLAS_LIB = path
                libraries = libs
                break
        else:
            continue
        break

    if ATLAS_LIB is None:
        print('Unable to find required libraries in the directory', ATLAS_LIB)
        print('Set SAGE_ATLAS_LIB to the directory containing:')
        print('- liblapack, libcblas, libf77blas and libatlas;')
        print('- or liblapack and libblas;')
        print('you wish to use existing ATLAS libraries.')
        print('For more details, see:')
        print('https://doc.sagemath.org/html/en/installation/source.html#environment-variables')
        print('Unset SAGE_ATLAS_LIB to build ATLAS from source.')
        print('Then type make.')
        sys.exit(2)

    for fname in os.listdir(ATLAS_LIB):
        if fname.startswith(prefix + 'f77blas'):
            f77blas = os.path.join(ATLAS_LIB, fname)
            break
    else:
        f77blas = None
    if f77blas is not None:
        symbol_table = try_run('readelf -s ' + f77blas)
    else:
        symbol_table = None
    if symbol_table is not None:
        sym_gfortran = 'gfortran' in symbol_table
        sym_g95 = 'g95' in symbol_table
        if sym_gfortran and conf['fortran'] != 'gfortran':
            print("Symbols in lib77blas indicate it was built with gfortran.\n")
            print("However, Sage is using a different fortran compiler.\n")
            print("If you wish to use this blas library, make sure FC points\n")
            print("to a fortran compiler compatible with this library.\n")
            sys.exit(2)
        if sym_g95 and conf['fortran'] != 'g95':
            print("Symbols in lib77blas indicate it was built with g95 \n")
            print("However, Sage is using a different fortran compiler.\n")
            print("If you wish to use this blas library, make sure FC points\n")
            print("to a fortran compiler compatible with this library.\n")
            sys.exit(2)

    SAGE_LOCAL_LIB = os.path.join(conf['SAGE_LOCAL'], 'lib')
    def symlinkOSlibrary(library_basename):
        filenames = [ fname for fname in os.listdir(path)
                      if fname.startswith(library_basename) ]
        for fname in filenames:
            source = os.path.join(ATLAS_LIB, fname)
            destination = os.path.join(SAGE_LOCAL_LIB, fname)
            print('Symlinking '+destination+' -> '+source)
            try:
                os.remove(destination)
            except OSError:
                pass
            try:
                os.symlink(source, destination)
            except OSError:
                pass
    for lib in libraries + libraries_optional:
        symlinkOSlibrary(prefix+lib)

    if 'atlas' in libraries:
        write_pc_file(['cblas', 'atlas'], 'cblas')
        write_pc_file(['f77blas', 'atlas'], 'blas')
        # The inclusion of cblas is not a mistake. ATLAS' lapack include
        # a custom version of clapack which depends on cblas.
        write_pc_file(['lapack', 'f77blas', 'cblas', 'atlas'], 'lapack')
    else:
        write_pc_file(['blas'], 'cblas')
        write_pc_file(['blas'], 'blas')
        write_pc_file(['lapack', 'blas'], 'lapack')


    sys.exit(0)

# Because blas, cblas and lapack libraries are properly linked
# with no unknown symbols, no extra libraries needs to be given.
write_pc_file(['cblas'], 'cblas')
write_pc_file(['f77blas'], 'blas')
write_pc_file(['lapack'], 'lapack')

# add extra architectural defaults
cp('src/ARCHS/*.tar.bz2', 'src/ATLAS/CONFIG/ARCHS/')

# hardcoded gcc in SpewMakeInc.c
edit_in_place('src/ATLAS/CONFIG/src/SpewMakeInc.c') \
    .replace('   goodgcc = .*', '   goodgcc = "' + os.environ['CC'] + '";') \
    .close()


# override throttling check if architecture is specified
edit_in_place('src/ATLAS/CONFIG/src/config.c') \
    .replace('if \(mach == MACHOther\)', 'if (mach != MACHOther) thrchk=0; else') \
    .close()



######################################################################
### configure functions
######################################################################

# For debug purposes one can install static libraries via SAGE_ATLAS_ARCH=static,...
INSTALL_STATIC_LIBRARIES = False

def configure_options_from_environment():
    # Figure out architecture (see ATLAS_MACHTYPE) and isa extensions (see
    # ATLAS_ISAEXT) from environment variables:
    arch = None
    isa_ext = None
    thread_limit = None
    if conf['generic_binary?']:
        print('Sage "fat" binary mode set: Building "base" binary')
        print('NOTE: This can result in a Sage that is significantly slower at certain numerical')
        print('linear algebra since full FAT binary support has not been implemented yet.')
        arch = 'generic'
    if 'SAGE_ATLAS_ARCH' not in os.environ:
        return (arch, isa_ext, thread_limit)
    for opt in os.environ['SAGE_ATLAS_ARCH'].split(','):
        if opt.startswith('thread'):
            thread_limit = int(opt.split(':')[-1])
        elif opt == 'static':
            global INSTALL_STATIC_LIBRARIES
            INSTALL_STATIC_LIBRARIES = True
        elif opt in ATLAS_MACHTYPE + ('fast', 'base'):
            if arch is not None:
                raise ValueError('multiple architectures specified: {0} and {1}'.format(arch, opt))
            arch = opt
        elif opt in ATLAS_ISAEXT:
            if isa_ext is None:
                isa_ext = [opt]
            else:
                isa_ext.append(opt)
        else:
            print('unknown SAGE_ATLAS_ARCH option: '+opt)
            print('SAGE_ATLAS_ARCH architecture must be "fast", "base", or one of '+ \
                ', '.join(ATLAS_MACHTYPE))
            print('SAGE_ATLAS_ARCH ISA extension must be one of '+ ', '.join(ATLAS_ISAEXT))
            sys.exit(1)
    return (arch, isa_ext, thread_limit)


def configure(arch=None, isa_ext=None):
    """
    Configure for ``arch``.

    INPUT:

    - ``arch`` -- ``None`` or one of ``ATLAS_MACHTYPE``

    - ``isa_ext`` -- ``None`` or a sublist of ``ATLAS_ISAEXT``

    OUTPUT: 0 if ``configure`` was successful, 1 if it failed.
    """
    try:
        if arch is None:
            arch, isa_ext, thread_limit = configure_options_from_environment()
        if arch == 'fast':
            arch, isa_ext, thread_limit = configure_fast()
        if arch == 'generic' or arch == 'base':
            arch, isa_ext, thread_limit = configure_base()
    except NotImplementedError:
        return 1

    print('Running configure with arch = '+str(arch)+ \
        ', isa extensions '+str(isa_ext), ' thread limit '+str(thread_limit))
    if os.path.isdir(BUILD_DIR):
        os.chdir(conf['SPKG_DIR'])   # Solaris/ZFS can't delete cwd
        shutil.rmtree(BUILD_DIR)
    os.mkdir(BUILD_DIR)
    os.chdir(BUILD_DIR)

    # We need to provide full pathes to FC and CC to ATLAS configure script,
    # so that it does not use 'find' and travel around the filesystem to find
    # them.
    # We first split the compiler executable names from potential options, e.g.
    # as in 'gcc -m64', then use 'which' to locate them, and finally add the
    # options back.
    FC = os.environ['FC']
    FCsplit = FC.find(' ')
    if  FCsplit != -1:
        FCbin, FCopt = FC[:FCsplit], FC[FCsplit:]
    else:
        FCbin, FCopt = FC, ''

    CC = os.environ['CC']
    CCsplit = CC.find(' ')
    if  CCsplit != -1:
        CCbin, CCopt = CC[:CCsplit], CC[CCsplit:]
    else:
        CCbin, CCopt = CC, ''

    cmd = '../ATLAS/configure'
    cmd += ' --prefix=' + conf['SAGE_LOCAL']
    cmd += ' --with-netlib-lapack-tarfile=' + LAPACK_TARFILE
    cmd += ' --cc="' + CC + '"'

    ## -Si latune 1: enable lapack tuning
    ## typically adds 3-4 hours of install time
    cmd += ' -Si latune 0'

    # Set flags for all compilers so we can build dynamic libraries
    ALLFLAGS = "-fPIC " + os.environ["LDFLAGS"]
    cmd += ' -Fa alg "{}"'.format(ALLFLAGS)

    # set number of threads limit: 0 = off, -1 = unlimited
    if thread_limit is not None:
        cmd += ' -t ' + str(thread_limit)

    # set fortran compiler
    cmd += ' -C if "' + which(FCbin) + FCopt + '"'

    # set C compiler
    cmd += ' -C acg "' + which(CCbin) + CCopt + '"'

    # set bit width
    cmd += ' -b ' + conf['bits'][0:2]

    # set OS type
    try:
        if conf['Darwin?']:
            atlas_osnam = 'OSX'
        elif conf['CYGWIN?']:
# Picking Win64 currently does not work on Cygwin though it might be a better
# choice in the future.
# Not that ATLAS does not seem to officialy support Cygwin64 anyway in 3.10.1.
#            if conf['64bit?']:
#                atlas_osnam = 'Win64'
#            else:
            atlas_osnam = 'WinNT'
        else:
            atlas_osnam = conf['system']
        atlas_system = ATLAS_OSTYPE.index(atlas_osnam)
        cmd += ' -O '+str(atlas_system)
    except ValueError:
        pass

    # use hard floats on ARM
    if conf['ARM?']:
        cmd += ' -D c -DATL_ARM_HARDFP=1'

    # set machine architecture
    if arch is not None:
        cmd += ' -A '+str(ATLAS_MACHTYPE.index(arch))

    # set cpu instruction set extensions
    if isa_ext is not None:
        isa_extension = sum(1 << ATLAS_ISAEXT.index(x) for x in isa_ext)
        cmd += ' -V '+str(isa_extension)

    # Custom configure options
    if conf['user']:
        cmd += " " + conf['user']

    rc = system_with_flush(cmd)
    make_check_enums()
    return rc


def configure_fast():
    isa_ext = ('None',)
    thread_limit = None
    if conf['Intel?'] and conf['64bit?']:
        print('Fast configuration on Intel x86_64 compatible CPUs.')
        arch = 'P4E'
        if not conf['CYGWIN?']: # cannot use assembly on Cygwin64
            isa_ext = ('SSE3', 'SSE2', 'SSE1')
    elif conf['Intel?'] and conf['32bit?']:
        print('Fast configuration on Intel i386 compatible CPUs.')
        arch = 'x86SSE2'
        isa_ext = ('SSE2', 'SSE1')
    elif conf['SPARC?']:
        print('Fast configuration on SPARC.')
        arch = 'USIV'
    elif conf['PPC?']:
        print('Fast configuration on PPC.')
        arch = 'PPCG5'
        isa_ext = ('AltiVec', )
    elif conf['IA64?']:
        print('Fast configuration on Itanium.')
        arch = 'IA64Itan2'
    elif conf['ARM?']:
        print('Fast configuration on ARM.')
        arch='ARMv7'
    else:
        raise NotImplementedError('I don\'t know a "fast" configuration for your cpu.')
    return (arch, isa_ext, thread_limit)


def configure_base():
    isa_ext = ('None',)
    thread_limit = 0   # disable threading in "base"
    if conf['Intel?'] and conf['64bit?']:
        print('Generic configuration on Intel x86_64 compatible CPUs.')
        arch = 'x86SSE2'
        if not conf['CYGWIN?']: # cannot use assembly on Cygwin64
            isa_ext = ('SSE2', 'SSE1')
    elif conf['Intel?'] and conf['32bit?']:
        print('Generic configuration on Intel i386 compatible CPUs.')
        arch = 'x86x87'
    elif conf['SPARC?']:
        print('Base configuration on SPARC.')
        arch = 'USIII'
    elif conf['PPC?']:
        print('Base configuration on PPC.')
        arch = 'PPCG4'
    elif conf['IA64?']:
        print('Base configuration on Itanium.')
        arch = 'IA64Itan'
    elif conf['ARM?']:
        print('Base configuration on ARM.')
        arch = 'ARMv6'
    else:
        raise NotImplementedError('I don\'t know a "base" configuration for your cpu.')
    return (arch, isa_ext, thread_limit)


######################################################################
### make function
######################################################################

def make_atlas(target=None):
    os.chdir(BUILD_DIR)
    if target is None:
        return system_with_flush(MAKE)
    else:
        return system_with_flush(MAKE + ' ' + target)


def make_atlas_library(target=None):
    os.chdir(os.path.join(BUILD_DIR, 'lib'))
    cmd = (MAKE + ' ' + target) if target else MAKE
    return system_with_flush(cmd)


######################################################################
### make and save archdef function
######################################################################

def build_and_save_archdef():
    try:
        ARCHDEF_SAVE_DIR = os.environ['SAGE_ATLAS_SAVE_ARCHDEF']
    except KeyError:
        return
    os.chdir(os.path.join(BUILD_DIR, 'ARCHS'))
    rc = system_with_flush(MAKE + ' ArchNew')
    assert_success(rc, bad='Making archdef failed.',
                   good='Finished building archdef.')
    rc = system_with_flush(MAKE + ' tarfile')
    assert_success(rc, bad='Making archdef tarfile failed.',
                   good='Finished building archdef tarfile.')
    for tarfile in glob.glob('*.tar.bz2'):
        cp(tarfile, ARCHDEF_SAVE_DIR)


######################################################################
### static libraries functions
######################################################################

def build(arch=None, isa_ext=None):
    """
    Configure/build with given architectural information.

    Return ``0`` if successfull.
    """
    rc = configure(arch, isa_ext)
    if rc:
        print("Configure failed.")
        return rc
    print("Finished configuring ATLAS.")
    return make_atlas()

def build_tuning():
    """
    Configure/build by going through the full tuning process.

    Return ``0`` if successfull.
    """
    rc = configure()
    if rc!=0:
        print('Configure failed, possibly because you have CPU throttling enabled.')
        print('Skipping tuning attempts.')
        return rc
    print('First attempt: automatic tuning.')
    rc = make_atlas()
    if rc==0:
        return rc
    print('ATLAS failed to build, possibly because of throttling or a loaded system.')
    print('Waiting 1 minute...')
    sys.stdout.flush()
    time.sleep(60)
    print('Second attempt: Re-running make.')
    return make_atlas()


######################################################################
### shared library hack functions
######################################################################

def configure_shared_library():
    os.chdir(BUILD_LIB_DIR)
    static_library_dir = os.path.join(BUILD_DIR, 'lib')
    for static_lib in glob.glob(os.path.join(static_library_dir, 'lib*.a')):
        shutil.copy(static_lib, BUILD_LIB_DIR)
    cmd = './configure'
    cmd += ' --prefix=' + conf['SAGE_LOCAL']
    cmd += ' --libdir=' + os.path.join(conf['SAGE_LOCAL'],'lib')
    cmd += ' --disable-static'
    return system_with_flush(cmd)

def make_shared_library(target=None):
    os.chdir(BUILD_LIB_DIR)
    cmd = (MAKE + ' ' + target) if target else MAKE
    return system_with_flush(cmd)


######################################################################
### build atlas and lapack static libraries
######################################################################

#
# Workaround for specific platforms: Disable tuning and go straight to
# fast/base architectural defaults
#
skip_tuning = conf['IA64?']    # Itanium is dead and tuning is broken


rc = None
if 'SAGE_ATLAS_ARCH' in os.environ or conf['generic_binary?']:
    print('Building using specific architecture.')
    rc = build()
else:
    print('Configuring ATLAS.')
    if skip_tuning:
        print('Skipping tuning attempts (skip_tuning = True).')
        rc = 1    # Fake failed tuning attempts
    else:
        rc = build_tuning()
    if rc!=0:
        print('Third attempt: use "fast" options.')
        rc = build(arch='fast')
    if rc!=0:
        print('Fourth attempt: use "base" options.')
        rc = build(arch='base')

assert_success(rc, bad='Failed to build ATLAS.', good='Finished building ATLAS core.')

build_and_save_archdef()


######################################################################
### build ATLAS shared libraries
######################################################################

rc = make_atlas_library('shared')
if rc!=0:
    print('Failed to build shared library (using the ATLAS build system)')
else:
    print('Installed ATLAS shared library (ATLAS build system)')

rc = make_atlas_library('ptshared')
if rc!=0:
    print('Failed to build threaded library (using the ATLAS build system)')
else:
    print('Installed ATLAS multi-threaded shared library (ATLAS build system)')


######################################################################
### configure and build atlas and lapack shared library hack
######################################################################

rc = configure_shared_library()
assert_success(rc, bad='Configuring shared ATLAS library failed (libtool).',
               good='Finished configuring shared ATLAS library (libtool).')

have_serial_libs = False
have_parallel_libs = False

rc = make_shared_library()
if rc!=0:
    print('Failed to build serial+parallel shared libraries, possibly because your')
    print('system does not support both. Trying to build serial libraries only (libtool).')
    rc = make_shared_library('all_serial')
    if rc!=0:
        print('Failed to build any shared library, installing static library as last resort (libtool).')
        INSTALL_STATIC_LIBRARIES = True
    else:
        print('Finished building serial shared ATLAS library (libtool).')
        have_serial_libs = True
else:
    have_parallel_libs = True


######################################################################
### install shared library hack
######################################################################

if have_parallel_libs:
    rc = make_shared_library('install')
    assert_success(rc, bad='Installing the parallel+serial shared ATLAS library failed (libtool).',
                   good='Finished installing parallel+serial shared ATLAS library (libtool).')

if have_serial_libs:
    rc = make_shared_library('install_serial')
    assert_success(rc, bad='Installing the serial shared ATLAS library failed (libtool).',
                   good='Finished installing serial shared ATLAS library (libtool).')


######################################################################
### install atlas and lapack headers and libraries
######################################################################

if not INSTALL_STATIC_LIBRARIES:
    edit_in_place(os.path.join(BUILD_DIR, 'Make.top')) \
        .replace('.*/liblapack.a.*', '') \
        .replace('.*/libcblas.a.*', '') \
        .replace('.*/libf77blas.a.*', '') \
        .replace('.*/libptcblas.a.*', '') \
        .replace('.*/libptf77blas.a.*', '') \
        .replace('.*/libatlas.a.*', '') \
        .close()

rc = make_atlas('install')
assert_success(rc, bad='Failed to install ATLAS headers',
               good='Installed ATLAS headers')


######################################################################
### install script to tune and build ATLAS
######################################################################

cp(os.path.join(PATCH_DIR, 'atlas-config'),
    os.path.join(conf['SAGE_LOCAL'], 'bin'))



# Import setuptools before importing distutils, so that setuptools
# can replace distutils by its own vendored copy.
import setuptools

from distutils import log
from distutils.command.build import build

class sage_build(build):
    sub_commands = [('build_cython', lambda *args: True)] + build.sub_commands

    def run_autogen(self):
        """
        Generate auto-generated sources.

        This must be done before building the python modules,
        see :trac:`22106`.
        """
        from sage_setup.autogen import autogen_all
        from sage_setup.find import find_python_sources
        from sage.env import SAGE_SRC

        log.info("Generating auto-generated sources")

        pkgs = autogen_all()
        python_packages, python_modules, cython_modules = find_python_sources(
            SAGE_SRC, [ pkg.replace('.', '/') for pkg in pkgs])

        for pkg in python_packages:
            if pkg not in self.distribution.packages:
                self.distribution.packages.append(pkg)

        for cython_module in cython_modules:
            self.distribution.ext_modules.append(cython_module)

    def run(self):
        self.run_autogen()
        build.run(self)



import os
import sys
import glob
from subprocess import check_call, CalledProcessError

try:
    SAGE_LOCAL = os.environ['SAGE_LOCAL']
    DOT_SAGE = os.environ['DOT_SAGE']
except KeyError:
    print('You need to run this script in a Sage shell ("sage -sh")')
    sys.exit(1)


TEXLIVE_INSTALL_UNIX_URL = \
    'http://mirror.ctan.org/systems/texlive/tlnet/install-tl-unx.tar.gz'

TEXLIVE_PROFILE_TEMPLATE = \
"""
selected_scheme scheme-minimal
TEXDIR {SAGE_LOCAL}/share/texlive
TEXMFLOCAL {SAGE_LOCAL}/share/texlive/texmf-local
TEXMFSYSCONFIG {SAGE_LOCAL}/share/texlive/texmf-config
TEXMFSYSVAR {SAGE_LOCAL}/share/texlive/texmf-var
TEXMFCONFIG {DOT_SAGE}/texlive/texmf-config
TEXMFHOME {DOT_SAGE}/texlive/texmf-home
TEXMFVAR {DOT_SAGE}/texlive/texmf-var
collection-basic 1
in_place 0
option_adjustrepo 1
option_autobackup 1
option_backupdir tlpkg/backups
option_desktop_integration 1
option_doc 1
option_file_assocs 1
option_fmt 1
option_letter 0
option_menu_integration 1
option_path 1
option_post_code 1
option_src 1
option_sys_bin {SAGE_LOCAL}/bin
option_sys_info {SAGE_LOCAL}/share/info
option_sys_man {SAGE_LOCAL}/share/man
option_w32_multi_user 1
option_write18_restricted 1
portable 0
"""


def have_texlive():
    try:
        check_call(['tlmgr', '--version'])
        return True
    except (OSError, CalledProcessError):
        return False


def download_install_script(tarball):
    try:
        from urllib import urlretrieve
    except ImportError:
        from urllib.request import urlretrieve
    urlretrieve(TEXLIVE_INSTALL_UNIX_URL, tarball)

    
def write_profile(filename):
    profile = TEXLIVE_PROFILE_TEMPLATE.format(
        SAGE_LOCAL=SAGE_LOCAL, DOT_SAGE=DOT_SAGE)
    with open(filename, 'w') as f:
        f.write(profile)
    print('TeXlive unattended install profile: {0}'.format(filename))

def first_dir(glob_pattern):
    """
    Return the first directory matching ``glob_pattern``
    """
    for dirent in glob.glob(glob_pattern):
        if os.path.isdir(dirent):
            return dirent
    raise RuntimeError('no directory found: {0}'.format(glob_pattern))
    
def install_texlive():
    import tempfile
    tmp_dir = tempfile.mkdtemp()
    tarball = os.path.join(tmp_dir, 'install-tl-unx.tar.gz')
    profile = os.path.join(tmp_dir, 'sage.profile')
    download_install_script(tarball)
    write_profile(profile)

    original_dir = os.getcwd()
    os.chdir(tmp_dir)
    check_call(['tar', '-x', '-z', '-f', tarball])
    install_dir = first_dir('install-tl-*')
    os.chdir(install_dir)

    check_call([
        './install-tl',
        '-profile',
        profile
    ])
    os.chdir(original_dir)
    

def install_packages():
    package_list_txt = os.path.join(
        os.path.dirname(__file__),
        'package-list.txt'
    )
    with open(package_list_txt) as f:
        package_list = f.read()
    packages = []
    for pkg in package_list.splitlines():
        pkg = pkg.strip()
        if len(pkg) == 0:
            continue
        if pkg.startswith('#'):
            continue
        packages.append(pkg)
    print('installing the following TeXlive packages:')
    for pkg in packages:
        print(' *  ' + pkg)
    check_call(['tlmgr', 'install'] + packages)
    check_call(['tlmgr', 'path', 'add'])
        

if __name__ == '__main__':
    if have_texlive():
        print('Using your own texlive install (see "tlmgr --version")')
    else:
        print('Performing minimal texlive install into {0}'
              .format(SAGE_LOCAL))
        install_texlive()
    install_packages()

